#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import functools
import itertools

import networkx

from pareidoscope.utils import nx_graph

import json
from networkx.readwrite import json_graph


def get_subgraphs_nx(query_graph, target_graph, vertex_candidates=None):
    """Return a list of subgraphs of target_graph that are isomorphic to
    query_graph

    """
    bfo_graph, bfo_to_raw = get_bfo(target_graph)
    if vertex_candidates is None:
        vertex_candidates = nx_graph.get_vertex_candidates(query_graph, bfo_graph)
    else:
        # convert candidates to bfo_graph
        raw_to_bfo = {v: k for k, v in bfo_to_raw.items()}
        vertex_candidates = [set([raw_to_bfo[vc] for vc in vcs]) for vcs in vertex_candidates]
    if any([len(vc) == 0 for vc in vertex_candidates]):
        return
    vertex_candidates = functools.reduce(lambda x, y: x.union(y), vertex_candidates)
    for subgraph in enumerate_connected_subgraphs(bfo_graph, bfo_to_raw, query_graph.number_of_nodes(), query_graph.number_of_edges(), vertex_candidates):
        vc = nx_graph.get_vertex_candidates(query_graph, subgraph)
        if match_yes_no(query_graph, subgraph, vc, 0):
            yield subgraph


def get_subgraphs(query_graph, target_graph):
    """Return a list of subgraphs of target_graph that are isomorphic
    to query_graph

    Arguments:
    - `query_graph`: query graph as adjacency matrix
    - `target_graph`: target graph as adjacency matrix
    """
    query_graph = nx_graph.create_nx_digraph(query_graph)
    target_graph = nx_graph.create_nx_digraph(target_graph)
    return get_subgraphs_nx(query_graph, target_graph)


def subsumes(query_graph, target_graph):
    """Return whether the query_graph subsumes the target graph."""
    qg = nx_graph.create_nx_digraph(query_graph)
    tg = nx_graph.create_nx_digraph(target_graph)
    return subsumes_nx(qg, tg)


def subsumes_nx(query_graph, target_graph, vertex_candidates=None):
    """Return whether the query_graph subsumes the target graph."""
    qg = query_graph
    tg = target_graph
    if qg.number_of_nodes() > tg.number_of_nodes():
        return False
    if qg.number_of_edges() > tg.number_of_edges():
        return False
    if vertex_candidates is None:
        vertex_candidates = nx_graph.get_vertex_candidates(qg, tg)
    if any([len(cands) == 0 for cands in vertex_candidates]):
        return False
    return match_yes_no(qg, tg, vertex_candidates, 0)


def get_choke_point_matches(query_graph, target_graph, choke_point, vertex_candidates=None):
    """Return the vertices from target that match the choke point vertex
    from query.

    Arguments:
        query_graph:
        target_graph:
        choke_point:

    Returns:
        Vertices from target_graph that correspond to the choke point
        vertex from query_graph.

    """
    if vertex_candidates is None:
        vertex_candidates = nx_graph.get_vertex_candidates(query_graph, target_graph)
    if any([len(vc) == 0 for vc in vertex_candidates]):
        return
    for choke_point_candidate in vertex_candidates[choke_point]:
        local_candidates = copy.deepcopy(vertex_candidates)
        local_candidates[choke_point] = set([choke_point_candidate])
        if subsumes_nx(query_graph, target_graph, local_candidates):
            yield choke_point_candidate


def choke_point_subsumes_nx(query_graph, target_graph, choke_point, choke_point_candidate, vertex_candidates=None):
    """Return whether the query_graph subsumes the target graph."""
    if query_graph.number_of_nodes() > target_graph.number_of_nodes():
        return False
    if query_graph.number_of_edges() > target_graph.number_of_edges():
        return False
    if vertex_candidates is None:
        vertex_candidates = nx_graph.get_vertex_candidates(query_graph, target_graph)
    else:
        vertex_candidates = copy.deepcopy(vertex_candidates)
    vertex_candidates[choke_point] = vertex_candidates[choke_point].intersection(set([choke_point_candidate]))
    if any([len(cands) == 0 for cands in vertex_candidates]):
        return False
    return match_yes_no(query_graph, target_graph, vertex_candidates, 0)


def get_bfo(target_graph, fragment=False):
    """Return target_graph in breadth-first-order as well as a mapping of
    vertices.

    Arguments:
    - `target_graph`:
    - `fragment`: is target graph a fragment?

    Raises:
    - IndexError: No root vertex has been found

    """
    graph = networkx.DiGraph()
    agenda = []
    vertex_counter = 1
    raw_to_bfo, bfo_to_raw = {}, {}
    roots = sorted([v[0] for v in target_graph.nodes(data=True) if "root" in v[1]])
    if len(roots) == 0:
        assert(fragment)
        all_vertices = set(target_graph.nodes())
        roots = sorted([v for v in target_graph.nodes() if all([networkx.has_path(target_graph, v, u) for u in all_vertices - set([v])])])
        if len(roots) == 0:
            roots = sorted([v for v in target_graph.nodes() if target_graph.in_degree(v) == 0])
            if len(roots) > 1:
                agenda = roots[1:]
                for v in agenda:
                    raw_to_bfo[v] = vertex_counter
                    bfo_to_raw[vertex_counter] = v
                    vertex_counter += 1
    root = roots[0]
    raw_to_bfo[root] = 0
    bfo_to_raw[0] = root
    graph.add_node(0, target_graph.node[root])
    agenda.insert(0, root)
    seen_vertices = set([])
    while len(agenda) > 0:
        vertex = agenda.pop(0)
        if vertex in seen_vertices:
            continue
        seen_vertices.add(vertex)
        edges = target_graph.out_edges(nbunch=[vertex], data=True)
        edges.sort(key=lambda x: (x[2]["relation"], x[1]))
        for source, target, label in edges:
            if target not in raw_to_bfo:
                raw_to_bfo[target] = vertex_counter
                bfo_to_raw[vertex_counter] = target
                agenda.append(target)
                vertex_counter += 1
            graph.add_node(raw_to_bfo[target], target_graph.node[target])
            graph.add_edge(raw_to_bfo[vertex], raw_to_bfo[target], label)
    return graph, bfo_to_raw


def enumerate_connected_subgraphs(graph, graph_to_raw, nr_of_vertices, nr_of_edges, vertex_candidates):
    """Python reimplementation of my implementation of the following
    algorithm
    (https://mpi-inf.mpg.de/departments/d5/teaching/ss09/queryoptimization/lecture8.pdf):

    EnumerateCsg(G)
    for all i ∈ [n − 1, ... , 0] descending {
        emit {v_i};
        EnumerateCsgRec(G , {v_i}, B_i );
    }

    EnumerateCsgRec(G, S, X)
    N = N(S) \ X;
    for all S' ⊆ N, S' ≠ ∅, enumerate subsets first {
        emit (S ∪ S');
    }
    for all S' ⊆ N, S' ≠ ∅, enumerate subsets first {
        EnumerateCsgRec(G, (S ∪ S'), (X ∪ N));
    }

    Perl implementation:
    trunk/ResourcesPareidoscope/Create/hpc_01_collect_dependency_subgraphs.pl

    """
    for vertex in sorted(graph.nodes(), reverse=True):
        max_subgraph_size = len([v for v in graph.nodes() if v >= vertex])
        if nr_of_vertices > max_subgraph_size:
            continue
        if vertex not in vertex_candidates:
            continue
        subgraph = networkx.DiGraph()
        #subgraph = [[{} for j in range(max_subgraph_size)] for i in range(max_subgraph_size)]
        subgraph.add_node(vertex, graph.node[vertex])
        nr_of_subgraph_vertices = subgraph.number_of_nodes()
        nr_of_subgraph_edges = subgraph.number_of_edges()
        if nr_of_subgraph_vertices == nr_of_vertices and nr_of_edges == nr_of_subgraph_edges:
            yield return_corpus_order(subgraph, graph_to_raw)
            continue
        prohibited_edges = set([])
        for v in [u for u in graph.nodes() if u < vertex]:
            prohibited_edges.update(set(graph.out_edges(nbunch=[v])))
            prohibited_edges.update(set(graph.in_edges(nbunch=[v])))
        # prohibit edges that do not connect to vertices from vertex_candidates
        for v in [u for u in graph.nodes() if u >= vertex]:
            prohibited_edges.update(set([e for e in graph.out_edges(nbunch=[v]) if e[1] not in vertex_candidates]))
            prohibited_edges.update(set([e for e in graph.in_edges(nbunch=[v]) if e[0] not in vertex_candidates]))
        for result in enumerate_connected_subgraphs_recursive(graph, subgraph, prohibited_edges, nr_of_vertices, nr_of_edges, graph_to_raw):
            yield result


def enumerate_connected_subgraphs_recursive(graph, subgraph, prohibited_edges, nr_of_vertices, nr_of_edges, graph_to_raw):
    """See enumerate_connected_subgraphs()"""
    out_edges, in_edges, neighbours = set([]), set([]), set([])
    neighbour_edges = {}
    subgraph_size = subgraph.number_of_nodes()
    subgraph_edges = subgraph.number_of_edges()
    for vertex in subgraph.nodes():
        for e in graph.out_edges(nbunch=[vertex]):
            if e in prohibited_edges:
                continue
            out_edges.add(e)
            neighbours.add(e[1])
            if e[1] not in neighbour_edges:
                neighbour_edges[e[1]] = []
            neighbour_edges[e[1]].append(e)
        for e in graph.in_edges(nbunch=[vertex]):
            if e in prohibited_edges:
                continue
            in_edges.add(e)
            neighbours.add(e[0])
            if e[0] not in neighbour_edges:
                neighbour_edges[e[0]] = []
            neighbour_edges[e[0]].append(e)
    # all combinations of vertices
    for combination in powerset(neighbours, 1, nr_of_vertices - subgraph_size):
        # all combinations of edges
        edges = tuple([powerset(neighbour_edges[vertex], 1, nr_of_edges - subgraph_edges) for vertex in combination])
        for edge_combi in itertools.product(*edges):
            ec = []
            for edges in edge_combi:
                ec.extend(edges)
            # new vertices
            new_vertices = set([e[0] if e[0] in neighbours else e[1] for e in ec])
            if len(new_vertices) > nr_of_vertices - subgraph_size:
                continue
            # edges between new vertices
            new_edges = set([])
            for new_vertex in new_vertices:
                new_edges.update([e for e in graph.out_edges(nbunch=[new_vertex]) if e[1] in new_vertices])
            for new_edge_combi in powerset(new_edges, 0, nr_of_edges - (subgraph_edges + len(ec))):
                nec = list(new_edge_combi)
                local_subgraph = subgraph.copy()
                # add edges
                for s, t in ec + nec:
                    local_subgraph.add_edge(s, t, graph.edge[s][t])
                    for k, v in graph.node[s].items():
                        local_subgraph.node[s][k] = v
                    for k, v in graph.node[t].items():
                        local_subgraph.node[t][k] = v
                nr_of_subgraph_vertices = local_subgraph.number_of_nodes()
                nr_of_subgraph_edges = local_subgraph.number_of_edges()
                if nr_of_subgraph_edges != subgraph_edges + len(ec + nec):
                    raise Exception("WAAAAH")
                if nr_of_subgraph_vertices == nr_of_vertices and nr_of_edges == nr_of_subgraph_edges:
                    yield return_corpus_order(local_subgraph, graph_to_raw)
                elif nr_of_subgraph_vertices < nr_of_vertices:
                    for result in enumerate_connected_subgraphs_recursive(graph, local_subgraph, prohibited_edges.union(in_edges, out_edges, new_edges), nr_of_vertices, nr_of_edges, graph_to_raw):
                        yield result


def enumerate_csg_minmax(graph, graph_to_raw, min_vertices=2, max_vertices=5):
    """Based on
    https://mpi-inf.mpg.de/departments/d5/teaching/ss09/queryoptimization/lecture8.pdf:

    EnumerateCsg(G)
    for all i ∈ [n − 1, ... , 0] descending {
        emit {v_i};
        EnumerateCsgRec(G , {v_i}, B_i );
    }

    EnumerateCsgRec(G, S, X)
    N = N(S) \ X;
    for all S' ⊆ N, S' ≠ ∅, enumerate subsets first {
        emit (S ∪ S');
    }
    for all S' ⊆ N, S' ≠ ∅, enumerate subsets first {
        EnumerateCsgRec(G, (S ∪ S'), (X ∪ N));
    }

    """
    for vertex in sorted(graph.nodes(), reverse=True):
        subgraph = networkx.DiGraph()
        subgraph.add_node(vertex, graph.node[vertex])
        nr_of_subgraph_vertices = subgraph.number_of_nodes()
        if min_vertices <= nr_of_subgraph_vertices <= max_vertices:
            yield return_corpus_order(subgraph, graph_to_raw)
        prohibited_edges = set([])
        for v in [u for u in graph.nodes() if u < vertex]:
            prohibited_edges.update(set(graph.out_edges(nbunch=[v])))
            prohibited_edges.update(set(graph.in_edges(nbunch=[v])))
        for result in enumerate_csg_minmax_recursive(graph, subgraph, prohibited_edges, graph_to_raw, min_vertices, max_vertices):
            yield result


def enumerate_csg_minmax_recursive(graph, subgraph, prohibited_edges, graph_to_raw, min_vertices, max_vertices):
    """See enumerate_connected_subgraphs()"""
    out_edges, in_edges, neighbours = set([]), set([]), set([])
    neighbour_edges = {}
    subgraph_edges = subgraph.number_of_edges()
    for vertex in subgraph.nodes():
        for e in graph.out_edges(nbunch=[vertex]):
            if e in prohibited_edges:
                continue
            out_edges.add(e)
            neighbours.add(e[1])
            if e[1] not in neighbour_edges:
                neighbour_edges[e[1]] = []
            neighbour_edges[e[1]].append(e)
        for e in graph.in_edges(nbunch=[vertex]):
            if e in prohibited_edges:
                continue
            in_edges.add(e)
            neighbours.add(e[0])
            if e[0] not in neighbour_edges:
                neighbour_edges[e[0]] = []
            neighbour_edges[e[0]].append(e)
    # all combinations of vertices
    for combination in powerset(neighbours, 1):
        # all combinations of edges
        edges = tuple([powerset(neighbour_edges[vertex], 1) for vertex in combination])
        for edge_combi in itertools.product(*edges):
            ec = []
            for edges in edge_combi:
                ec.extend(edges)
            # new vertices
            new_vertices = set([e[0] if e[0] in neighbours else e[1] for e in ec])
            # edges between new vertices
            new_edges = set([])
            for new_vertex in new_vertices:
                new_edges.update([e for e in graph.out_edges(nbunch=[new_vertex]) if e[1] in new_vertices])
            for new_edge_combi in powerset(new_edges, 0):
                nec = list(new_edge_combi)
                local_subgraph = subgraph.copy()
                # add edges
                for s, t in ec + nec:
                    local_subgraph.add_edge(s, t, graph.edge[s][t])
                    for k, v in graph.node[s].items():
                        local_subgraph.node[s][k] = v
                    for k, v in graph.node[t].items():
                        local_subgraph.node[t][k] = v
                nr_of_subgraph_vertices = local_subgraph.number_of_nodes()
                nr_of_subgraph_edges = local_subgraph.number_of_edges()
                if nr_of_subgraph_edges != subgraph_edges + len(ec + nec):
                    raise Exception("WAAAAH")
                if min_vertices <= nr_of_subgraph_vertices <= max_vertices:
                    yield return_corpus_order(local_subgraph, graph_to_raw)
                if nr_of_subgraph_vertices < max_vertices:
                    for result in enumerate_csg_minmax_recursive(graph, local_subgraph, prohibited_edges.union(in_edges, out_edges, new_edges), graph_to_raw, min_vertices, max_vertices):
                        yield result


def enumerate_csg_minmax_node_induced(graph, graph_to_raw, min_vertices=2, max_vertices=5):
    """Enumerate all connected subgraphs of graph with at least
    min_vertices and at most max_vertices that are induced from a
    subset of its vertices (i.e. that include all edges between the
    vertices).

    """
    for vertex in sorted(graph.nodes(), reverse=True):
        vertices = set([vertex])
        nr_of_subgraph_vertices = len(vertices)
        if min_vertices <= nr_of_subgraph_vertices <= max_vertices:
            subgraph = graph.subgraph(vertices)
            yield return_corpus_order(subgraph, graph_to_raw)
        prohibited_vertices = set([v for v in graph.nodes() if v <= vertex])
        for result in enumerate_csg_minmax_node_induced_recursive(graph, vertices, prohibited_vertices, graph_to_raw, min_vertices, max_vertices):
            yield result


def enumerate_csg_minmax_node_induced_recursive(graph, vertices, prohibited_vertices, graph_to_raw, min_vertices, max_vertices):
    """See enumerate_connected_subgraphs()"""
    neighbours = set()
    for vertex in vertices:
        neighbours.update(set(graph.successors(vertex)))
        neighbours.update(set(graph.predecessors(vertex)))
    neighbours.difference_update(prohibited_vertices | vertices)
    for subset in powerset(neighbours, 1):
        local_vertices = vertices | set(subset)
        nr_of_subgraph_vertices = len(local_vertices)
        if min_vertices <= nr_of_subgraph_vertices <= max_vertices:
            local_subgraph = graph.subgraph(local_vertices)
            yield return_corpus_order(local_subgraph, graph_to_raw)
        if nr_of_subgraph_vertices < max_vertices:
            for result in enumerate_csg_minmax_node_induced_recursive(graph, local_vertices, prohibited_vertices | neighbours, graph_to_raw, min_vertices, max_vertices):
                yield result


def powerset(iterable, min_size, max_size=-1):
    """(Partial) powerset of set with at most max_size elements."""
    s = list(iterable)
    if max_size == -1:
        max_size = len(s)
    max_size = min(max_size, len(s))
    #return itertools.chain(itertools.combinations(s, r) for r in range(max_size + 1))
    powerset = []
    for r in range(min_size, max_size + 1):
        powerset.extend(list(itertools.combinations(s, r)))
    return powerset


def return_corpus_order(subgraph, bfo_to_raw):
    """"""
    corpus_order = networkx.DiGraph()
    corpus_order.add_nodes_from([(bfo_to_raw[v], l) for v, l in subgraph.nodes(data=True)])
    corpus_order.add_edges_from([(bfo_to_raw[s], bfo_to_raw[t], l) for s, t, l in subgraph.edges(data=True)])
    return corpus_order


def match_yes_no(query_graph, target_graph, vertex_candidates, index):
    """Return wether there is at least one subgraph isomorphism between
    query_graph and target_graph.

    """
    if index >= query_graph.number_of_nodes():
        return True
    query_outgoing = query_graph.out_edges(nbunch=[index], data=True)
    query_incoming = query_graph.in_edges(nbunch=[index], data=True)
    for cpos in vertex_candidates[index]:
        local_candidates = [cand - set([cpos]) for cand in vertex_candidates]
        local_candidates[index] = set([cpos])
        target_outgoing = target_graph.out_edges(nbunch=[cpos], data=True)
        target_incoming = target_graph.in_edges(nbunch=[cpos], data=True)
        target_candidates = [set([]) for _ in query_graph.nodes()]
        for qs, qt, ql in query_outgoing:
            for ts, tt, tl in target_outgoing:
                if nx_graph.dictionary_match(ql, tl):
                    target_candidates[qt] = target_candidates[qt].union(set([tt]))
        for qs, qt, ql in query_incoming:
            for ts, tt, tl in target_incoming:
                if nx_graph.dictionary_match(ql, tl):
                    target_candidates[qs] = target_candidates[qs].union(set([ts]))
        for idx in range(query_graph.number_of_nodes()):
            if len(target_candidates[idx]) > 0:
                local_candidates[idx] = local_candidates[idx].intersection(target_candidates[idx])
        if any([len(x) == 0 for x in local_candidates]):
            continue
        local_result = match_yes_no(query_graph, target_graph, local_candidates, index + 1)
        if local_result is True:
            return True
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import re

import networkx


def create_nx_digraph(adj_matrix):
    """Return a networkx.DiGraph object of the adjacency matrix."""
    dg = networkx.DiGraph()
    dg.add_nodes_from([(i, adj_matrix[i][i]) for i in range(len(adj_matrix))])
    for i in range(len(adj_matrix)):
        for j in range(len(adj_matrix)):
            if i == j:
                continue
            if adj_matrix[i][j]:
                dg.add_edge(i, j, adj_matrix[i][j])
    return dg


def create_nx_digraph_from_cwb(cwb, origid=None):
    """Return a networkx.DiGraph object of the CWB representation."""
    relpattern = re.compile(r"^(?P<relation>[^(]+)[(](?P<offset>-?\d+)(?:&apos;|')*,0(?:&apos;|')*[)]$")
    dg = networkx.DiGraph()
    if origid is not None:
        dg.graph["origid"] = origid
    attributes = lambda l: {"word": l[0], "pos": l[1], "lemma": l[2], "wc": l[3]}
    dg.add_nodes_from([(l, attributes(cwb[l])) for l in range(len(cwb))])
    for l in range(len(cwb)):
        if cwb[l][6] == "root":
            dg.node[l]["root"] = "root"
    for i, line in enumerate(cwb):
        indeps = line[4]
        if indeps != "|":
            for rel in indeps.strip("|").split("|"):
                match = re.search(relpattern, rel)
                relation = match.group("relation")
                offset = int(match.group("offset"))
                # ignore dependency relations for punctuation
                if relation == "punct":
                    continue
                if offset != 0:
                    dg.add_edge(i + offset, i, {"relation": relation})
    # remove unconnected vertices, e.g. punctuation in the SD model
    for v, l in dg.nodes(data=True):
        if "root" not in l and dg.degree(v) == 0:
            dg.remove_node(v)
    # make sure that the remaining vertices are consecutively labeled
    dg = ensure_consecutive_vertices(dg)
    return dg


def create_nx_digraph_from_conllu(conllu, origid=None):
    """Return a networkx.DiGraph object of the CoNLL-U representation."""
    dg = networkx.DiGraph()
    if origid is not None:
        dg.graph["origid"] = origid
    attributes = lambda l: {"word": l[1], "lemma": l[2], "wc": l[3], "pos": l[4]}
    dg.add_nodes_from([(l, attributes(conllu[l])) for l in range(len(conllu))])
    id_to_enumeration = {conllu[l][0]: l for l in range(len(conllu))}
    for l in range(len(conllu)):
        if conllu[l][7] == "root":
            dg.node[l]["root"] = "root"
    for i, line in enumerate(conllu):
        relations = set()
        if line[7] != "root" and line[7] != "_":
            relations.add((id_to_enumeration[line[6]], line[7]))
        if line[8] != "_":
            for rel in line[8].split("|"):
                gov, relation = rel.split(":", maxsplit=1)
                governor = id_to_enumeration[gov]
                if relation != "root":
                    relations.add((governor, relation))
        for governor, relation in relations:
            if relation == "root" or relation == "punct":
                continue
            dg.add_edge(governor, i, {"relation": relation})
    # remove unconnected vertices, e.g. punctuation in the SD model
    for v, l in dg.nodes(data=True):
        if "root" not in l and dg.degree(v) == 0:
            dg.remove_node(v)
    # make sure that the remaining vertices are consecutively labeled
    dg = ensure_consecutive_vertices(dg)
    return dg


def is_sensible_graph(nx_graph):
    """Check if graph is a sensible syntactic representation of a
    sentence, i.e. rooted, connected, sensible outdegree, sensible
    density, …

    """
    # are the vertices consecutively labeled?
    if sorted(nx_graph.nodes()) != list(range(nx_graph.number_of_nodes())):
        raise Exception("Vertices are not consecutively labeled")
    # is there a vertex explicitly labeled as "root"?
    roots = [v for v, l in nx_graph.nodes(data=True) if "root" in l]
    if len(roots) != 1:
        return False
    # is the graph connected?
    if not networkx.is_weakly_connected(nx_graph):
        return False
    # is the "root" vertex really a root, i.e. is there a path to
    # every other vertex?
    root = roots[0]
    other_vertices = set(nx_graph.nodes())
    other_vertices.remove(root)
    if not all([networkx.has_path(nx_graph, root, v) for v in other_vertices]):
        return False
    # # do the vertices have sensible outdegrees <= 10?
    if any([nx_graph.out_degree(v) > 10 for v in nx_graph.nodes()]):
        return False
    # is the graph overly dense, i.e. is there a vertex with an
    # extended star (neighbors + edges between them) with more than 18
    # edges?
    if any([nx_graph.subgraph(set([v] + nx_graph.predecessors(v) + nx_graph.successors(v))).number_of_edges() > 18 for v in nx_graph.nodes()]):
        return False
    return True


def get_vertex_candidates(query_graph, target_graph):
    """For each vertex in query_graph return a list of possibly
    corresponding vertices from target_graph.

    """
    candidates = [[] for v in query_graph.nodes()]
    for vertex in query_graph.nodes():
        if len(query_graph.node[vertex]) == 0:
            candidates[vertex] = target_graph.nodes()
        else:
            candidates[vertex] = [tv for tv in target_graph.nodes() if dictionary_match(query_graph.node[vertex], target_graph.node[tv])]
        query_in = query_graph.in_degree(vertex)
        query_out = query_graph.out_degree(vertex)
        candidates[vertex] = [tv for tv in candidates[vertex] if (query_in <= target_graph.in_degree(tv)) and (query_out <= target_graph.out_degree(tv))]
        # negated edges
        not_indep = query_graph.node[vertex].get("not_indep")
        if not_indep is not None:
            not_indep = set(not_indep)
            candidates[vertex] = [tv for tv in candidates[vertex] if len(not_indep & set([l["relation"] for s, t, l in target_graph.in_edges(tv, data=True)])) == 0]
        not_outdep = query_graph.node[vertex].get("not_outdep")
        if not_outdep is not None:
            not_outdep = set(not_outdep)
            candidates[vertex] = [tv for tv in candidates[vertex] if len(not_outdep & set([l["relation"] for s, t, l in target_graph.out_edges(tv, data=True)])) == 0]
        candidates[vertex] = set(candidates[vertex])
    # verify adjacency and edge labels
    for s, t, l in query_graph.edges(data=True):
        source_candidates = set([])
        target_candidates = set([])
        for cs, ct in itertools.product(candidates[s], candidates[t]):
            if target_graph.has_edge(cs, ct):
                if dictionary_match(l, target_graph.edge[cs][ct]):
                    source_candidates.add(cs)
                    target_candidates.add(ct)
        candidates[s] = source_candidates
        candidates[t] = target_candidates
    return candidates


def dictionary_match(query_dictionary, target_dictionary):
    """Does target_vertex match query_vertex? I.e. are all keys from
    query_vertex in target_vertex and if yes, are their values equal?

    """
    results = []
    for label in query_dictionary:
        if label.startswith("not_"):
            if label == "not_indep" or label == "not_outdep":
                continue
            pos_label = label[4:]
            results.append(pos_label in target_dictionary and query_dictionary[label] != target_dictionary[pos_label])
        else:
            results.append(label in target_dictionary and query_dictionary[label] == target_dictionary[label])
    return all(results)
    # return all([label in target_dictionary and (query_dictionary[label] == target_dictionary[label] or re.search(r"^" + str(query_dictionary[label]) + r"$", str(target_dictionary[label]))) for label in query_dictionary])


def export_to_adjacency_matrix(nx_graph, canonical=False):
    """Return an adjacency matrix representing nx_graph."""
    # identity mapping
    mapping = {v: v for v in nx_graph.nodes()}
    # or mapping to canonical order
    if canonical:
        mapping = {v: i for i, v in enumerate(canonical_order(nx_graph))}
    matrix = [[{} for j in range(nx_graph.number_of_nodes())] for i in range(nx_graph.number_of_nodes())]
    for v, l in nx_graph.nodes(data=True):
        vm = mapping[v]
        matrix[vm][vm] = l
    for s, t, l in nx_graph.edges(data=True):
        sm = mapping[s]
        tm = mapping[t]
        matrix[sm][tm] = l
    return matrix


def export_to_cwb_format(nx_graph):
    """Do actual formatting work."""
    output = []
    for v in sorted(nx_graph.nodes()):
        word = nx_graph.node[v]["word"]
        pos = nx_graph.node[v].get("pos", "X")
        lemma = nx_graph.node[v].get("lemma", word)
        wc = nx_graph.node[v].get("wc", "X")
        root = nx_graph.node[v].get("root", "")
        indeps = ["%s(%d,0)" % (l["relation"], s - v) for s, t, l in nx_graph.in_edges(v, data=True)]
        outdeps = ["%s(0,%d)" % (l["relation"], t - v) for s, t, l in nx_graph.out_edges(v, data=True)]
        indeps = "|" + "|".join(indeps)
        if len(indeps) > 1:
            indeps += "|"
        outdeps = "|" + "|".join(outdeps)
        if len(outdeps) > 1:
            outdeps += "|"
        output.append([word, pos, lemma, wc, indeps, outdeps, root])
    return output


def export_to_chemical_format(graph, vertex_dict, edge_dict):
    """Export to chemical format for use with Gaston
    (http://www.liacs.nl/~snijssen/gaston/index.html).

    """
    export = ["t # "]
    vmax = None
    try:
        vmax = max(vertex_dict.values())
    except ValueError:
        vmax = 0
    for v in sorted(graph.nodes()):
        word = graph.node[v]["word"]
        pos = graph.node[v].get("pos", "X")
        lemma = graph.node[v].get("lemma", word)
        wc = graph.node[v].get("wc", "X")
        t = (word, pos, lemma, wc)
        if t not in vertex_dict:
            vmax += 1
            vertex_dict[t] = vmax
        export.append("v %d %d" % (v, vertex_dict[t]))
        emax = None
    try:
        emax = max(edge_dict.values())
    except ValueError:
        emax = 0
    for s, t, l in sorted(graph.edges(data=True)):
        rel = l["relation"]
        if rel not in edge_dict:
            emax += 1
            edge_dict[rel] = emax
        export.append("e %d %d %d" % (s, t, edge_dict[rel]))
    return export


def is_purely_structural(nx_graph):
    """Is query_graph purely structural, i.e. does it contain no
    restrictions on vertex and edge labels?

    """
    vertices = all([_is_unrestricted(l) for v, l in nx_graph.nodes(data=True)])
    edges = all([_is_unrestricted(l) for s, t, l in nx_graph.edges(data=True)])
    return vertices and edges


def _is_unrestricted(dictionary):
    """Is the dictionary unrestricted, i.e. is it empty or does it only
    have ".+" or ".*" as values?

    """
    if dictionary == {}:
        return True
    return all([val == ".+" or val == ".*" for val in dictionary.values()])


def _get_vertex_tuple(nx_graph, vertex):
    """Return a tuple for the vertex that can be used for sorting"""
    other_vertices = set(nx_graph.nodes())
    other_vertices.remove(vertex)
    root = all((networkx.has_path(nx_graph, vertex, x) for x in other_vertices))
    antiroot = all((networkx.has_path(nx_graph, x, vertex) for x in other_vertices))
    star_center = all((nx_graph.has_edge(vertex, x) or nx_graph.has_edge(x, vertex) for x in other_vertices))
    choke_point = None
    if root or antiroot or star_center:
        choke_point = True
    else:
        choke_point = all((networkx.has_path(nx_graph, vertex, x) or networkx.has_path(nx_graph, x, vertex) for x in other_vertices))
    indegree = nx_graph.in_degree(vertex)
    outdegree = nx_graph.out_degree(vertex)
    label = tuple(sorted(nx_graph.node[vertex].items()))
    inedgelabels = tuple(sorted([tuple(sorted(nx_graph.edge[s][t].items())) for s, t in nx_graph.in_edges(vertex)]))
    outedgelabels = tuple(sorted([tuple(sorted(nx_graph.edge[s][t].items())) for s, t in nx_graph.out_edges(vertex)]))
    return (root, antiroot, star_center, choke_point, indegree, outdegree, label, inedgelabels, outedgelabels)


def _dfs(nx_graph, vertex, vtuples, return_ids=False, blacklist=[]):
    """Return vertex tuples in order of depth-first search starting from
    vertex

    """
    order = []
    seen = set(blacklist)
    agenda = [vertex]
    keyfunc = lambda v: vtuples[v]
    while len(agenda) > 0:
        v = agenda.pop(0)
        if v in seen:
            continue
        seen.add(v)
        if return_ids:
            order.append(v)
        else:
            order.append(vtuples[v])
        successors = sorted([x for x in nx_graph.successors(v) if x not in seen], key=keyfunc)
        agenda = _get_unique_order(nx_graph, successors, vtuples, blacklist=list(seen)) + agenda
    return order


def _get_unique_order(nx_graph, sorted_vertices, vtuples, blacklist=[]):
    """Resolve any groups within sorted_vertices"""
    order = []
    keyfunc = lambda v: vtuples[v]
    for k, g in itertools.groupby(sorted_vertices, keyfunc):
        group = list(g)
        if len(group) == 1:
            order.append(group[0])
        elif len(group) == 0:
            raise Exception("Group should not be empty!")
        else:
            # We need more criteria for sorting
            vtuples_dfs = {v: tuple(_dfs(nx_graph, v, vtuples, blacklist=blacklist)) for v in group}
            keyfunc_dfs = lambda v: (vtuples_dfs[v], v)
            order.extend(sorted(group, key=keyfunc_dfs))
    return order


def canonical_order(nx_graph):
    """Return the vertices of nx_graph in canonical order"""
    vertices = nx_graph.nodes()
    vtuples = {v: _get_vertex_tuple(nx_graph, v) for v in vertices}
    roots = [v for v in vtuples if vtuples[v][0]]
    antiroots = [v for v in vtuples if vtuples[v][1]]
    if len(roots) == 1:
        return _dfs(nx_graph, roots[0], vtuples, return_ids=True)
    if len(antiroots) == 1:
        return list(reversed(_dfs(nx_graph.reverse(), antiroots[0], vtuples, return_ids=True)))
    keyfunc = lambda v: vtuples[v]
    sorted_vertices = sorted(vertices, key=keyfunc)
    return _get_unique_order(nx_graph, sorted_vertices, vtuples)


def canonize(nx_graph, order=False):
    """Return canonized copy of nx_graph"""
    co = canonical_order(nx_graph)
    mapping = {v: i for i, v in enumerate(co)}
    canonized = networkx.DiGraph(**nx_graph.graph)
    for v, l in nx_graph.nodes(data=True):
        canonized.add_node(mapping[v], l)
    for s, t, l in nx_graph.edges(data=True):
        canonized.add_edge(mapping[s], mapping[t], l)
    if order:
        return canonized, co
    else:
        return canonized


def skeletize(nx_graph, only_vertices=False):
    """Return skeleton copy of nx_graph"""
    skeleton = networkx.DiGraph(**nx_graph.graph)
    for s, t, l in nx_graph.edges(data=True):
        if only_vertices:
            skeleton.add_edge(s, t, {k: v for k, v in l.items()})
        else:
            skeleton.add_edge(s, t)
    return skeleton


def get_choke_point(nx_graph):
    """Return one of the choke point vertices if there are any.

    Arguments:
        nx_graph:

    Returns:
        The first root vertex in canonical order. If there aren't
        any, the first antiroot vertex in canonical order. Else the
        first other choke point in canonical order.

    """
    vertices = canonical_order(nx_graph)
    vtuples = {v: _get_vertex_tuple(nx_graph, v) for v in vertices}
    roots = [v for v in vertices if vtuples[v][0]]
    antiroots = [v for v in vertices if vtuples[v][1]]
    choke_points = [v for v in vertices if vtuples[v][2]]
    if len(roots) >= 1:
        return roots[0]
    if len(antiroots) >= 1:
        return antiroots[0]
    if len(choke_points) >= 1:
        return choke_points[0]
    return None


def is_star(nx_graph, return_centers=False):
    """Check if the graph is a star, i.e. if it has one vertex that is
    adjacent to all others.

    """
    vertices = set(nx_graph.nodes())
    if return_centers:
        centers = [v for v in vertices if set([v] + nx_graph.predecessors(v) + nx_graph.successors(v)) == vertices]
        is_star = len(centers) >= 1
        return is_star, set(centers)
    else:
        return any((set([v] + nx_graph.predecessors(v) + nx_graph.successors(v)) == vertices for v in vertices))


def is_star_center(nx_graph, v):
    """Check if the graph is a star and the vertex v is its center, i.e.
    if v is adjacent to all other vertices.

    """
    vertices = set(nx_graph.nodes())
    return set([v] + nx_graph.predecessors(v) + nx_graph.successors(v)) == vertices


def ensure_consecutive_vertices(graph):
    """Make sure that the n vertices of the graph have indices 0 to
    n-1.

    """
    if sorted(graph.nodes()) == list(range(graph.number_of_nodes())):
        return graph
    else:
        mapping = {v: i for i, v in enumerate(sorted(graph.nodes()))}
        new_graph = networkx.DiGraph(**graph.graph)
        for vertex, label in graph.nodes(data=True):
            new_graph.add_node(mapping[vertex], label)
        for s, t, l in graph.edges(data=True):
            new_graph.add_edge(mapping[s], mapping[t], l)
        return new_graph



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json

import networkx
from networkx.readwrite import json_graph

from pareidoscope import subgraph_enumeration
from pareidoscope import subgraph_isomorphism
from pareidoscope.utils import nx_graph


def _extract_stars(graph, vertex, edge_stars, skel_stars, edge_to_skel={}, include_pos=False):
    """Extract star-like subgraphs from sentence for a given center
    vertex

    Arguments:
    - `graph`:
    - `position`:

    """
    identity_mapping = {v: v for v in graph.nodes()}
    pos = graph.node[vertex]["pos"]
    wc = graph.node[vertex]["wc"]
    whole_star = graph.subgraph(set([vertex] + graph.predecessors(vertex) + graph.successors(vertex)))
    for subgraph in subgraph_enumeration.enumerate_csg_minmax(whole_star, identity_mapping, min_vertices=2, max_vertices=whole_star.__len__()):
        # vertex must be in subgraph
        if not subgraph.has_node(vertex):
            continue
        # vertex must be center of star
        other_vertices = set(subgraph.nodes()) - set([vertex])
        if len(other_vertices) == 0:
            continue
        if not all((subgraph.has_edge(vertex, x) or subgraph.has_edge(x, vertex) for x in other_vertices)):
            continue
        length = subgraph.__len__()
        degree_sequence = " ".join([str(_) for _ in sorted([subgraph.degree(_) for _ in subgraph.nodes()], reverse=True)])
        sorted_edges = " ".join(sorted([l["relation"] for s, t, l in subgraph.edges(data=True)]))
        edge_star = nx_graph.skeletize(subgraph, only_vertices=True)
        edge_star, order = nx_graph.canonize(edge_star, order=True)
        edge_center = order.index(vertex)
        edge_t = tuple(networkx.generate_edgelist(edge_star, data=["relation"]))
        edge_string = json.dumps(list(edge_t), ensure_ascii=False, sort_keys=True)
        edge_star_key = (edge_string, edge_center, pos, wc) if include_pos else (edge_string, edge_center)
        if edge_star_key not in edge_stars:
            edge_stars[edge_star_key] = {"length": length, "degree_sequence": degree_sequence, "sorted_edges": sorted_edges, "star_freq": 1, "center_set": set([vertex])}
        else:
            edge_stars[edge_star_key]["star_freq"] += 1
            edge_stars[edge_star_key]["center_set"].add(vertex)
        skel_star = nx_graph.skeletize(subgraph)
        skel_star, order = nx_graph.canonize(skel_star, order=True)
        skel_center = order.index(vertex)
        skel_t = tuple(networkx.generate_edgelist(skel_star))
        skel_string = json.dumps(list(skel_t), ensure_ascii=False, sort_keys=True)
        if (skel_string, skel_center) not in skel_stars:
            skel_stars[(skel_string, skel_center)] = {"length": length, "degree_sequence": degree_sequence, "star_freq": 1, "center_set": set([vertex])}
        else:
            skel_stars[(skel_string, skel_center)]["star_freq"] += 1
            skel_stars[(skel_string, skel_center)]["center_set"].add(vertex)
        if (edge_string, edge_center) not in edge_to_skel:
            edge_to_skel[(edge_string, edge_center)] = set([(skel_string, skel_center)])
        else:
            edge_to_skel[(edge_string, edge_center)].add((skel_string, skel_center))
    return


def _center_sets_to_center_freqs(edge_stars, skel_stars):
    """Change center_set to center_freq in both dictionaries"""
    for e in edge_stars:
        edge_stars[e]["center_freq"] = len(edge_stars[e]["center_set"])
        edge_stars[e].pop("center_set")
    for s in skel_stars:
        skel_stars[s]["center_freq"] = len(skel_stars[s]["center_set"])
        skel_stars[s].pop("center_set")


def count_isomorphisms(query, target, query_vertex, target_vertex):
    """Count isomorphisms which map query_vertex to target_vertex"""
    candidates = nx_graph.get_vertex_candidates(query, target)
    candidates[query_vertex] = set([target_vertex])
    return len(list(subgraph_isomorphism.get_subgraph_isomorphisms_nx(query, target, candidates)))


def extract_stars_for_position(args):
    """Extract star-like subgraphs from sentence for a given position
    (token)

    Arguments:
    - `sentid`:
    - `graph`:
    - `position`:

    """
    sentid, graph, position = args
    edge_stars = {}
    skel_stars = {}
    edge_to_skel = {}
    gs = json_graph.node_link_graph(json.loads(graph))
    bfo_graph, bfo_to_raw = subgraph_enumeration.get_bfo(gs)
    raw_to_bfo = {v: k for k, v in bfo_to_raw.items()}
    vertex = raw_to_bfo[position]
    _extract_stars(bfo_graph, vertex, edge_stars, skel_stars, edge_to_skel, include_pos=False)
    _center_sets_to_center_freqs(edge_stars, skel_stars)
    return sentid, edge_stars, skel_stars, edge_to_skel


def extract_all_stars(args):
    """Extract all star-like subgraphs from sentence

    Arguments:
    - `sentid`:
    - `graph`:

    """
    sentid, graph = args
    edge_stars = {}
    skel_stars = {}
    gs = json_graph.node_link_graph(json.loads(graph))
    bfo_graph, bfo_to_raw = subgraph_enumeration.get_bfo(gs)
    for vertex in bfo_graph.nodes():
        _extract_stars(bfo_graph, vertex, edge_stars, skel_stars, include_pos=True)
    _center_sets_to_center_freqs(edge_stars, skel_stars)
    return sentid, edge_stars, skel_stars



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import multiprocessing

def _unpack_args(function):
    """Wrapper around function that accepts arguments as a single
    tuple."""
    def function_taking_tuple(arguments):
        function(*arguments)
    return function_taking_tuple


def map_reduce(map_function, reduce_function, map_arguments, reduce_arguments, processes=multiprocessing.cpu_count(), groupsize_multiplier=500, chunksize=10):
    """Parallel processing of the elements in arguments and reduction to a
    single value.
    
    Arguments:
        map_function:
        reduce_function:
        map_arguments: An iterable
        reduce_arguments: Additional arguments to reduce_function,
            passed as first arguments
        processes: Number of processes; default: number of cpus
        groupsize_multiplier: How many arguments per process in one go
        chunksize: 

    Returns:
        The result of reduce(reduce_function, reduce_arguments, map(map_function, map_arguments)).

    """
    groupsize = processes * groupsize_multiplier
    pool = multiprocessing.Pool(processes=processes)
    
    map_f = _unpack_args(map_function)
    reduce_f = functools.partial(reduce_function, *reduce_arguments)

    results = functools.reduce(reduce_f, pool.imap_unordered(map_f, arguments, chunksize=chunksize))
    return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import math
import random


def grouper_nofill(n, iterable):
    """list(grouper_nofill(3, 'ABCDEFG')) --> [['A', 'B', 'C'], ['D', 'E',
    'F'], ['G']]

    """
    it = iter(iterable)
    while True:
        group = list(itertools.islice(it, n))
        if not group:
            return
        yield group


def sample(iterable, p):
    """Return a sample of iterable where each element will be chosen with
    probability p.

    """
    return (e for e in iterable if random.random() < p)


def get_int_bins(min_value, max_value, nr_of_bins=10):
    """Find integer bin edges such that the bins are of equal size.

    Arguments:
        min_value:
        max_value:
        nr_of_bins:

    Returns:
        A list of bin edges.

    """
    bin_edges = None
    if max_value <= min_value + nr_of_bins - 1:
        bin_edges = list(range(min_value, max_value + 2))
    else:
        bin_size = int(math.ceil(float((max_value - min_value) + 1) / nr_of_bins))
        bin_edges = list(range(min_value, 1 + min_value + bin_size * nr_of_bins, bin_size))
    bin_edges[-1] = bin_edges[-1] - 1
    return bin_edges



#!/usr/bin/python
# -*- coding: utf-8 -*-

import math


def get_contingency_table(o11, r1, c1, n):
    """Fill in the values for O11, O12, O21 and O22 and E11, E12, E21 and
    E22.

    Args:
        o11:
        r1:
        c1:
        n:

    """
    o, e = {1: {}, 2: {}}, {1: {}, 2: {}}
    r2 = n - r1
    c2 = n - c1
    o[1][1] = float(o11)
    o[1][2] = float(r1 - o11)
    o[2][1] = float(c1 - o11)
    o[2][2] = float(r2 - o[2][1])
    e[1][1] = r1 * c1 / float(n)
    e[1][2] = r1 * c2 / float(n)
    e[2][1] = r2 * c1 / float(n)
    e[2][2] = r2 * c2 / float(n)
    return o, e


def poisson_stirling_log(o, e):
    """Calculate the Poisson-Stirling measure

    Args:
        o:
        e:

    """
    return o[1][1] * (math.log(o[1][1], 2) - math.log(e[1][1], 2) - 1)


def z_score(o, e):
    """Calculate the z-score measure

    Args:
        o:
        e:

    """
    return (o[1][1] - e[1][1]) / math.sqrt(e[1][1])


def t_score(o, e):
    """Calculate the t-score measure

    Args:
        o:
        e:

    """
    return (o[1][1] - e[1][1]) / math.sqrt(o[1][1])


def chi_squared(o, e):
    """Calculate the chi-squared measure

    Args:
        o:
        e:

    """
    n = o[1][1] + o[1][2] + o[2][1] + o[2][2]
    return (n * (o[1][1] - e[1][1]) ** 2) / (e[1][1] * e[2][2])


def one_sided_chi_squared(o, e):
    """Calculate the one-sided chi-squared measure

    Args:
        o:
        e:

    """

    if o[1][1] < e[1][1]:
        return -1.0 * chi_squared(o, e)
    else:
        return chi_squared(o, e)


def log_likelihood(o, e):
    """Calculate the log-likelihood measure

    Args:
        o:
        e:

    """
    ll = 0
    for r in o:
        for c in o[r]:
            if o[r][c] == 0:
                continue
            ll += o[r][c] * math.log(o[r][c] / e[r][c], 2)
    ll *= 2
    return ll


def one_sided_log_likelihood(o, e):
    """Calculate the one-sided log-likelihood measure

    Args:
        o:
        e:

    """
    if o[1][1] < e[1][1]:
        return -1.0 * log_likelihood(o, e)
    else:
        return log_likelihood(o, e)


def mutual_information(o, e):
    """Calculate the mutual information measure

    Args:
        o:
        e:

    """
    return math.log(o[1][1] / e[1][1], 2)


def local_mi(o, e):
    """Calculate the local mutual information measure

    Args:
        o:
        e:

    """
    return o[1][1] * mutual_information(o, e)


def average_mi(o, e):
    """Calculate the average mutual information measure (= log-likelihood)

    Args:
        o:
        e:

    """
    return log_likelihood(o, e)


def mi2(o, e):
    """Calculate the heuristic MI^2 measure

    Args:
        o:
        e:

    """
    return math.log(o[1][1] ** 2 / e[1][1], 2)


def mi3(o, e):
    """Calculate the heuristic MI^3 measure

    Args:
        o:
        e:

    """
    return math.log(o[1][1] ** 3 / e[1][1], 2)


def odds_ratio_disc(o, e):
    """Calculate the discounted odds-ratio measure

    Args:
        o:
        e:

    """
    return math.log(((o[1][1] + 0.5) * (o[2][2] + 0.5)) / ((o[1][2] + 0.5) * (o[2][1] + 0.5)), 2)


def relative_risk(o, e):
    """Calculate the relative-risk measure

    Args:
        o:
        e:

    """
    c1 = o[1][1] + o[2][1]
    c2 = o[1][2] + o[2][2]
    return math.log((o[1][1] * c2) / (o[1][2] * c1), 2)


def liddell(o, e):
    """Calculate the Liddell measure

    Args:
        o:
        e:

    """
    c1 = o[1][1] + o[2][1]
    c2 = o[1][2] + o[2][2]
    return (o[1][1] * o[2][2] - o[1][2] * o[2][1]) / (c1 * c2)


def minimum_sensitivity(o, e):
    """Calculate the minimum-sensitivity measure

    Args:
        o:
        e:

    """
    r1 = o[1][1] + o[1][2]
    c1 = o[1][1] + o[2][1]
    return min(o[1][1] / r1, o[1][1] / c1)


def geometric_mean(o, e):
    """Calculate the gmean measure

    Args:
        o:
        e:

    """
    r1 = o[1][1] + o[1][2]
    c1 = o[1][1] + o[2][1]
    return o[1][1] / math.sqrt(r1 * c1)


def dice(o, e):
    """Calculate the Dice measure

    Args:
        o:
        e:

    """
    r1 = o[1][1] + o[1][2]
    c1 = o[1][1] + o[2][1]
    return (2 * o[1][1]) / (r1 + c1)


def jaccard(o, e):
    """Calculate the Jaccard measure

    Args:
        o:
        e:

    """
    return o[1][1] / (o[1][1] + o[1][2] + o[2][1])


def cooccurrence_frequency(o, e):
    """Return O11

    Args:
        o:
        e:

    """
    return o[1][1]


def delta_p_column(o, e):
    """Calculate the column-wise delta P measure, i.e. the difference of
    column proportions (= Liddell).

    Args:
        o:
        e:

    """
    return liddell(o, e)


def delta_p_row(o, e):
    """Calculate the row-wise delta P measure, i.e. the difference of row
    proportions.

    Args:
        o:
        e:

    """
    r1 = o[1][1] + o[1][2]
    r2 = o[2][1] + o[2][2]
    return (o[1][1] * o[2][2] - o[1][2] * o[2][1]) / (r1 * r2)


def max_delta_p(o, e):
    """Return the maximum of the column-wise and the row-wise delta P
    measure.

    Args:
        o:
        e:

    """
    return max(delta_p_row(o, e), delta_p_column(o, e))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pareidoscope.utils import nx_graph
from pareidoscope import subgraph_enumeration
from pareidoscope import subgraph_isomorphism

def get_frequencies(args):
    """Determine various frequencies (isomorphisms, subgraphs, graphs) for
    the query structures.
    
    Arguments:
        args: Contains the arguments
    
    Returns:
        If a sentence is sensible: Number of matching isomorphisms,
        subgraphs and graphs.

    """
    s, query, root = args
    sentence, sid = s
    result = {}
    gs = nx_graph.create_nx_digraph_from_cwb(sentence)
    sensible = nx_graph.is_sensible_graph(gs)
    candidates = None
    if sensible:
        isomorphisms, subgraphs, roots, graphs = 0, 0, 0, 0
        if subgraph_enumeration.subsumes_nx(query, gs, vertex_candidates=candidates):
            graphs = 1
            isomorphisms = sum(1 for _ in subgraph_isomorphism.get_subgraph_isomorphisms_nx(query, gs, vertex_candidates=candidates))
            subgraphs = sum(1 for _ in subgraph_enumeration.get_subgraphs_nx(query, gs, vertex_candidates=candidates))
            if root is not None:
                roots = sum(1 for _ in subgraph_enumeration.get_root_matches(query, gs, root))
        result = {"isomorphisms": isomorphisms, "subgraphs": subgraphs, "roots": roots, "graphs": graphs}
    return sid, result, sensible


def merge_result(result, results):
    """Merge result with results
    
    Arguments:
    - `result`:
    - `results`:
    """
    for k in result:
        if k not in results:
            results[k] = 0
        results[k] += result[k]


def merge_nested_result(result, results):
    """Merge result with results
    
    Arguments:
    - `result`:
    - `results`:
    """
    for outer_key in result:
        if outer_key not in results:
            results[outer_key] = {}
        for inner_key, value in result[outer_key].items():
            if inner_key not in results[outer_key]:
                results[outer_key][inner_key] = 0
            results[outer_key][inner_key] += value



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import itertools
import json
import os
import random
import sys

import networkx

script_directory = os.path.abspath(os.path.dirname(sys.argv[0]))
if script_directory not in sys.path:
    sys.path.insert(0, script_directory)

import nx_graph


def get_random_near_tree(nr_of_vertices, vertex_label_distri, edge_label_distri, max_outdegree = 0):
    """Create a random near tree with nr_of_vertices vertices. The
    vertex and edge labels are determined by vertex_label_distri and
    edge_label_distri.
    
    Arguments:
    - `nr_of_vertices`:
    - `vertex_label_distri`:
    - `edge_label_distri`:
    - `max_outdegree`:
    """
    pass


def get_random_tree(nr_of_vertices, vertex_label_distri, edge_label_distri, degree_distri, max_outdegree=0, no_root=False, underspec=0.0):
    """Create a random tree with nr_of_vertices vertices. The vertex
    and edge labels are determined by vertex_label_distri and
    edge_label_distri.
    
    Arguments:
    - `nr_of_vertices`:
    - `vertex_label_distri`:
    - `edge_label_distri`:
    - `degree_distri`:
    - `max_outdegree`:
    """
    root = None
    directed_tree = networkx.DiGraph()
    if nr_of_vertices == 1:
        root = 0
        directed_tree.add_node(root)
    else:
        undirected_tree = create_tree_skeleton(nr_of_vertices, degree_distri, max_outdegree)
        root = random.choice(list(itertools.chain.from_iterable([[v] * undirected_tree.degree(v) for v in undirected_tree.nodes()])))
        directed_tree.add_edges_from(networkx.bfs_edges(undirected_tree, root))
    if not no_root:
        directed_tree.node[root]["root"] = "root"
    add_vertex_labels(directed_tree, vertex_label_distri, underspec)
    add_edge_labels(directed_tree, edge_label_distri, underspec)
    return directed_tree


def get_powerlaw_tree_sequence(nr_of_vertices):
    """Return a powerlaw degree sequence
    
    Arguments:
    - `nr_of_vertices`:
    """
    if nr_of_vertices == 1:
        return [0]
    try:
        degree_sequence = networkx.random_powerlaw_tree_sequence(nr_of_vertices, tries=1000)
    except networkx.NetworkXError:
        degree_sequence = get_powerlaw_tree_sequence(nr_of_vertices)
    return degree_sequence


def create_tree_skeleton(nr_of_vertices, degree_distri, max_outdegree):
    """Create a random undirected tree without any labels.
    
    Arguments:
    - `nr_of_vertices`:
    - `degree_distri`:
    - `max_outdegree`:
    """
    degree_sum = 2 * (nr_of_vertices - 1)
    degree_sequence = None
    if degree_distri == "uniform":
        degree_sequence = distribution(nr_of_vertices, degree_sum)
        if max_outdegree > 1:
            while max(degree_sequence) > max_outdegree:
                degree_sequence = distribution(nr_of_vertices, degree_sum)
    elif degree_distri == "zipf":
        mini = int(round(degree_sum * 0.75))
        maxi = int(round(degree_sum * 1.25))
        while True:
            degree_sequence = networkx.utils.zipf_sequence(nr_of_vertices)
            while sum(degree_sequence) < mini or sum(degree_sequence) > maxi:
                degree_sequence = networkx.utils.zipf_sequence(nr_of_vertices)
            x = [_ / float(sum(degree_sequence)) for _ in degree_sequence]
            while sum(degree_sequence) != degree_sum:
                degree_sequence = adjust_distribution(degree_sequence, x, degree_sum, nr_of_vertices - 1)
            if max(degree_sequence) <= max_outdegree:
                break
    elif degree_distri == "powerlaw":
        degree_sequence = get_powerlaw_tree_sequence(nr_of_vertices)
        if max_outdegree > 1:
            while max(degree_sequence) > max_outdegree:
                degree_sequence = get_powerlaw_tree_sequence(nr_of_vertices)
    undirected_tree = networkx.degree_sequence_tree(degree_sequence)
    return undirected_tree


def add_vertex_labels(tree, vertex_label_distri, underspec):
    """Add vertex labels to tree according to vertex_label_distri.
    
    Arguments:
    - `tree`:
    - `vertex_label_distri`:
    """
    for vertex in tree.nodes():
        if random.random() < underspec:
            tree.node[vertex]["word"] = ".+"
        else:
            tree.node[vertex]["word"] = random.choice(vertex_label_distri)


def add_edge_labels(tree, edge_label_distri, underspec):
    """Add vertex labels to tree according to vertex_label_distri.
    
    Arguments:
    - `tree`:
    - `vertex_label_distri`:
    """
    for s, t in tree.edges():
        if random.random() < underspec:
            tree.edge[s][t]["relation"] = ".+"
        else:
            tree.edge[s][t]["relation"] = random.choice(edge_label_distri)


def distribution(n, total):
    """Return n random natural numbers that sum up to total and follow
    a uniform distribution.
    
    Arguments:
    - `n`:
    - `total`:
    """
    maximum = total - (n - 1)
    p = [0] + sorted( random.random() for _ in range(n - 1) ) + [1]
    distri_1 = [p[i+1] - p[i] for i in range(n)]
    distri_total = [int(total * _) for _ in distri_1]
    # adjust lowest values
    distri_total = [max(1, _) for _ in distri_total]
    # adjust highest values
    distri_total = [min(maximum, _) for _ in distri_total]
    # adjust remaining values until total is correct
    while sum(distri_total) != total:
        distri_total = adjust_distribution(distri_total, distri_1, total, maximum)
    return distri_total


def adjust_distribution(dist_to_adjust, random_dist, total, maximum):
    """Adjust dist_to_adjust so that it sums up to total, no value is
    larger that maximum or smaller than one and relative difference to
    random_dist is minimal.
    
    Arguments:
    - `dist_to_adjust`:
    - `random_dist`:
    - `total`:
    - `maximum`:
    """
    deviations = [x - y for (x, y) in zip([_ / float(total) for _ in dist_to_adjust], random_dist)]
    deviations = list(enumerate(deviations))
    comp = maximum
    adj = 1
    if sum(dist_to_adjust) < total:
        deviations.sort(key = lambda x: x[1])
    else:
        deviations.sort(key = lambda x: x[1], reverse = True)
        comp = 1
        adj = -1
    for (idx, _) in deviations:
        if dist_to_adjust[idx] == comp:
            continue
        dist_to_adjust[idx] += adj
        break
    return dist_to_adjust


def create_distribution(size, distribution, prefix):
    """Create data according to distribution.
    
    Arguments:
    - `size`:
    - `distribution`:
    """
    vocabulary = []
    if distribution == "uniform":
        vocabulary = ["%s%05d" % (prefix, i) for i in range(size)]
    elif distribution == "zipf":
        zs = (int(round(z)) for z in networkx.utils.zipf_sequence(size))
        for i, j in enumerate(zs):
            vocabulary.extend(["%s%05d" % (prefix, i)] * j)
    return vocabulary


def get_length(sent_dist, mu, sigma, min_length, max_length):
    """Return an integer.
    
    Arguments:
    - `sent_dist`:
    - `mu`:
    - `sigma`:
    - `min_length`:
    - `max_length`:
    """
    length = None
    if sent_dist == "uniform":
        length = random.randint(min_length, max_length)
    elif sent_dist == "normal":
        length = int(round(random.normalvariate(mu, sigma)))
        while length < min_length or length > max_length:
            length = int(round(random.normalvariate(mu, sigma)))
    return length


def split_tree(tree):
    """Split tree into two subtrees and return queries.
    
    Arguments:
    - `tree`:
    """
    # choose one edge for splitting
    s, t = random.choice(tree.edges())
    t_bunch = set(networkx.dfs_preorder_nodes(tree, t))
    s_bunch = set(tree.nodes()) - t_bunch
    a = tree.subgraph(s_bunch)
    b = tree.subgraph(t_bunch)
    # give edge to either source or target vertex
    if random.choice(["source", "target"]) == "source":
        a.add_node(t, {"word": ".+"})
        a.add_edge(s, t, {"relation": tree[s][t]["relation"]})
    else:
        b.add_node(s, {"word": ".+"})
        b.add_edge(s, t, {"relation": tree[s][t]["relation"]})
    n = tree.copy()
    for vertex in n.nodes():
        n.node[vertex]["word"] = ".+"
    for s, t in n.edges():
        n.edge[s][t]["relation"] = ".+"
    r1 = n.copy()
    for vertex in a.nodes():
        r1.node[vertex]["word"] = a.node[vertex]["word"]
    for s, t in a.edges():
        r1.edge[s][t]["relation"] = a.edge[s][t]["relation"]
    c1 = n.copy()
    for vertex in b.nodes():
        c1.node[vertex]["word"] = b.node[vertex]["word"]
    for s, t in b.edges():
        c1.edge[s][t]["relation"] = b.edge[s][t]["relation"]
    return a, b, r1, c1, n


def cwb_format(tree, length, sid):
    """Do actual formatting work.
    
    Arguments:
    - `tree`:
    - `length`:
    """
    lines = []
    lines.append('<s id="%d" len="%d">' % (sid, length))
    lines.extend(["\t".join(l) for l in nx_graph.export_to_cwb_format(tree)])
    lines.append("</s>")
    return "\n".join(lines)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pareidoscope.utils import nx_graph


def get_subgraph_isomorphisms_nx(query_graph, target_graph, vertex_candidates=None):
    """Return a list of isomorphisms that map vertices and edges from
    query_graph to vertices and edges from target_graph.

    """
    subgraph_isomorphisms = None
    if vertex_candidates is None:
        vertex_candidates = nx_graph.get_vertex_candidates(query_graph, target_graph)
    if any([len(vc) == 0 for vc in vertex_candidates]):
        return iter(())
    if nx_graph.is_purely_structural(query_graph):
        subgraph_isomorphisms = match_structural(query_graph, target_graph, vertex_candidates, 0)
    else:
        subgraph_isomorphisms = match(query_graph, target_graph, vertex_candidates, 0)
    return subgraph_isomorphisms


def get_subgraph_isomorphisms(query_graph, target_graph):
    """Return a list of isomorphisms that map vertices and edges from
    query_graph to vertices and edges from target_graph.

    Arguments:
    - `query_graph`: query graph as adjacency matrix
    - `target_graph`: target graph as adjacency matrix
    """
    query_graph = nx_graph.create_nx_digraph(query_graph)
    target_graph = nx_graph.create_nx_digraph(target_graph)
    return get_subgraph_isomorphisms_nx(query_graph, target_graph)


def match(query_graph, target_graph, vertex_candidates, index):
    """Find isomorphisms between query_graph and target_graph.

    The implementation follows Proisl/Uhrig (2012: 2753) quite
    closely.

    """
    if index >= query_graph.number_of_nodes():
        yield tuple([next(iter(cand)) for cand in vertex_candidates])
    else:
        query_outgoing = query_graph.out_edges(nbunch=[index], data=True)
        query_incoming = query_graph.in_edges(nbunch=[index], data=True)
        for cpos in vertex_candidates[index]:
            local_candidates = [cand - set([cpos]) for cand in vertex_candidates]
            local_candidates[index] = set([cpos])
            target_outgoing = target_graph.out_edges(nbunch=[cpos], data=True)
            target_incoming = target_graph.in_edges(nbunch=[cpos], data=True)
            target_candidates = [set([]) for _ in query_graph.nodes()]
            for qs, qt, ql in query_outgoing:
                for ts, tt, tl in target_outgoing:
                    if nx_graph.dictionary_match(ql, tl):
                        target_candidates[qt] = target_candidates[qt].union(set([tt]))
            for qs, qt, ql in query_incoming:
                for ts, tt, tl in target_incoming:
                    if nx_graph.dictionary_match(ql, tl):
                        target_candidates[qs] = target_candidates[qs].union(set([ts]))
            for idx in range(query_graph.number_of_nodes()):
                if len(target_candidates[idx]) > 0:
                    local_candidates[idx] = local_candidates[idx].intersection(target_candidates[idx])
            if any([len(x) == 0 for x in local_candidates]):
                continue
            for result in match(query_graph, target_graph, local_candidates, index + 1):
                yield result


def match_structural(query_graph, target_graph, vertex_candidates, index):
    """Find isomorphisms between query_graph and target_graph.

    The implementation follows Proisl/Uhrig (2012: 2753) quite
    closely.

    """
    if index >= query_graph.number_of_nodes():
        yield tuple([next(iter(cand)) for cand in vertex_candidates])
    else:
        query_outgoing = query_graph.out_edges(nbunch=[index], data=True)
        query_incoming = query_graph.in_edges(nbunch=[index], data=True)
        for cpos in vertex_candidates[index]:
            local_candidates = [cand - set([cpos]) for cand in vertex_candidates]
            local_candidates[index] = set([cpos])
            target_outgoing = target_graph.out_edges(nbunch=[cpos], data=True)
            target_incoming = target_graph.in_edges(nbunch=[cpos], data=True)
            target_candidates = [set([]) for _ in query_graph.nodes()]
            tout_indexes = set([t for s, t, l in target_outgoing])
            tin_indexes = set([s for s, t, l in target_incoming])
            for s, t, l in query_outgoing:
                target_candidates[t] = target_candidates[t].union(tout_indexes)
            for s, t, l in query_incoming:
                target_candidates[s] = target_candidates[s].union(tin_indexes)
            for idx in range(query_graph.number_of_nodes()):
                if len(target_candidates[idx]) > 0:
                    local_candidates[idx] = local_candidates[idx].intersection(target_candidates[idx])
            if any([len(x) == 0 for x in local_candidates]):
                continue
            for result in match_structural(query_graph, target_graph, local_candidates, index + 1):
                yield result


if __name__ == "__main__":
    query_graph = [[{"word": "bachelor"}, {"relation": "amod"}], [{}, {}]]
    target_graph = [[{"word": "he"}, {}, {}, {}, {}, {}], [{}, {"word": "was"}, {}, {}, {}, {}], [{}, {}, {"word": "an"}, {}, {}, {}], [{}, {}, {}, {"word": "old"}, {}, {}], [{}, {}, {}, {}, {"word": "confirmed"}, {}], [{"relation": "nsubj"}, {"relation": "cop"}, {"relation": "det"}, {"relation": "amod"}, {"relation": "amod"}, {"word": "bachelor"}]]
    print(list(get_subgraph_isomorphisms(query_graph, target_graph)))
    query_graph = [[{}, {'relation': '.+'}, {'relation': '.+'}], [{}, {}, {}], [{}, {}, {}]]
    target_graph = [[{'root': 'root', 'word': 'v0989'}, {'relation': 'e00'}, {}, {}, {}, {}], [{}, {'word': 'v2248'}, {'relation': 'e29'}, {}, {'relation': 'e04'}, {}], [{}, {}, {'word': 'v5171'}, {'relation': 'e05'}, {}, {}], [{}, {}, {}, {'word': 'v7172'}, {}, {}], [{}, {}, {}, {}, {'word': 'v6658'}, {'relation': 'e20'}], [{}, {}, {}, {}, {}, {'word': 'v9605'}]]
    print(list(get_subgraph_isomorphisms(query_graph, target_graph)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import json

import networkx
from networkx.readwrite import json_graph

from pareidoscope.utils import nx_graph
from pareidoscope import subgraph_enumeration
from pareidoscope import subgraph_isomorphism


def read_queries(queries_file):
    """Read all queries."""
    queries = json.load(queries_file)
    for query in queries:
        graph = json_graph.node_link_graph(query, directed=True, multigraph=False)
        graph = nx_graph.ensure_consecutive_vertices(graph)
        for v, l in graph.nodes(data=True):
            l["vid"] = v
        yield graph


def strip_vid(graph):
    """Return a copy of graph without vertex id (vid) labels."""
    copy = graph.copy()
    for vertex in copy.nodes():
        del copy.node[vertex]["vid"]
    return copy


def _get_isomorphism_vertex_candidates(query_graph, normal_candidates, v_target, v_isomorphism, vid_to_iso):
    """Return vertex candidates that are compatible with isomorphism."""
    isomorphism_candidates = [set([vid_to_iso[l["vid"]]]) if l["vid"] in vid_to_iso else v_target - v_isomorphism for v, l in sorted(query_graph.nodes(data=True))]
    candidates = [a & b for a, b in zip(normal_candidates, isomorphism_candidates)]
    return candidates


def _get_subgraph_vertex_candidates(query_graph, normal_candidates, vid_n, v_target, v_subgraph):
    """Return vertex candidates that are compatible with subgraph."""
    subgraph_candidates = [v_subgraph if l["vid"] in vid_n else v_target - v_subgraph for v, l in sorted(query_graph.nodes(data=True))]
    candidates = [a & b for a, b in zip(normal_candidates, subgraph_candidates)]
    return candidates


def _frequency_signature(subsumes, args_c, args_a, args_b):
    """Return O11, R1, C1 and N for gs."""
    subsumed_by_gc = subsumes(**args_c)
    subsumed_by_ga = subsumes(**args_a)
    subsumed_by_gb = subsumes(**args_b)
    o11, r1, c1, n = 0, 0, 0, 0
    n = 1
    inconsistent = False
    if subsumed_by_gc and subsumed_by_ga and subsumed_by_gb:
        o11 = 1
        r1 = 1
        c1 = 1
    elif (not subsumed_by_gc) and subsumed_by_ga and (not subsumed_by_gb):
        r1 = 1
    elif (not subsumed_by_gc) and (not subsumed_by_ga) and subsumed_by_gb:
        c1 = 1
    elif (not subsumed_by_gc) and subsumed_by_ga and subsumed_by_gb:
        r1 = 0.5
        c1 = 0.5
        inconsistent = True
    elif (not subsumed_by_gc) and (not subsumed_by_ga) and (not subsumed_by_gb):
        pass
    else:
        raise Exception("Inconsistent classification.")
    return o11, r1, c1, n, inconsistent


def derive_subgraphs_and_focus_points(embeddings, focus_point_vertex):
    """"""
    subgraphs = set([frozenset(iso) for iso in embeddings])
    focus_points = set([iso[focus_point_vertex] for iso in embeddings])
    # embeddings, subgraphs, focus_points, sentences
    frequencies = (len(embeddings), len(subgraphs), len(focus_points), min(1, len(embeddings)))
    return subgraphs, focus_points, frequencies


def all_counting_methods(gc, ga, gb, gn, gs, focus_point, candidates=None):
    """Count embeddings and derive the other counting methods."""
    result = {}
    vs = set(gs.nodes())
    stripped_gc = strip_vid(gc)
    stripped_ga = strip_vid(ga)
    stripped_gb = strip_vid(gb)
    embeddings_gc, embeddings_ga, embeddings_gb = set(), set(), set()
    embeddings = set(subgraph_isomorphism.get_subgraph_isomorphisms_nx(strip_vid(gn), gs, vertex_candidates=candidates))
    if len(embeddings) > 0:
        normal_cand_c = nx_graph.get_vertex_candidates(stripped_gc, gs)
        normal_cand_a = nx_graph.get_vertex_candidates(stripped_ga, gs)
        normal_cand_b = nx_graph.get_vertex_candidates(stripped_gb, gs)
    for embedding in embeddings:
        vid_to_iso = {gn.node[qv]["vid"]: tv for qv, tv in zip(sorted(gn.nodes()), embedding)}
        v_embedding = set(embedding)
        vert_cand_c = _get_isomorphism_vertex_candidates(gc, normal_cand_c, vs, v_embedding, vid_to_iso)
        vert_cand_a = _get_isomorphism_vertex_candidates(ga, normal_cand_a, vs, v_embedding, vid_to_iso)
        vert_cand_b = _get_isomorphism_vertex_candidates(gb, normal_cand_b, vs, v_embedding, vid_to_iso)
        if subgraph_enumeration.subsumes_nx(stripped_gc, gs, vertex_candidates=vert_cand_c):
            embeddings_gc.add(embedding)
        if subgraph_enumeration.subsumes_nx(stripped_ga, gs, vertex_candidates=vert_cand_a):
            embeddings_ga.add(embedding)
        if subgraph_enumeration.subsumes_nx(stripped_gb, gs, vertex_candidates=vert_cand_b):
            embeddings_gb.add(embedding)
    subgraphs_n, focus_points_n, n = derive_subgraphs_and_focus_points(embeddings, focus_point)
    subgraphs_gc, focus_points_gc, o11 = derive_subgraphs_and_focus_points(embeddings_gc, focus_point)
    subgraphs_ga, focus_points_ga, r1 = derive_subgraphs_and_focus_points(embeddings_ga, focus_point)
    subgraphs_gb, focus_points_gb, c1 = derive_subgraphs_and_focus_points(embeddings_gb, focus_point)
    inconsistent_embeddings = (embeddings_ga & embeddings_gb) - embeddings_gc
    inconsistent_subgraphs = (subgraphs_ga & subgraphs_gb) - subgraphs_gc
    inconsistent_focus_points = (focus_points_ga & focus_points_gb) - focus_points_gc
    inconsistent_sentence = 1 if r1[3] == 1 and c1[3] == 1 and o11[3] == 0 else 0
    inconsistencies = (len(inconsistent_embeddings), len(inconsistent_subgraphs), len(inconsistent_focus_points), inconsistent_sentence)
    for i, cm in enumerate(("embeddings", "subgraphs", "focus_points", "sentences")):
        result[cm] = {"o11": o11[i], "r1": r1[i] - inconsistencies[i] / 2, "c1": c1[i] - inconsistencies[i] / 2, "n": n[i], "inconsistent": inconsistencies[i]}
    return result


def isomorphisms(gc, ga, gb, gn, gs, candidates=None):
    """Count isomorphisms

    Arguments:
    - `go11`:
    - `gr1`:
    - `gc1`:
    - `gn`:
    - `gs`:
    - `candidates`:

    """
    # vid_to_gn = {l["vid"]: v for v, l in gn.nodes(data=True)}
    ct = {x: 0 for x in ["o11", "r1", "c1", "n", "inconsistent"]}
    vs = set(gs.nodes())
    stripped_gc = strip_vid(gc)
    stripped_ga = strip_vid(ga)
    stripped_gb = strip_vid(gb)
    normal_cand_c, normal_cand_a, normal_cand_b = None, None, None
    for isomorphism in subgraph_isomorphism.get_subgraph_isomorphisms_nx(strip_vid(gn), gs, vertex_candidates=candidates):
        if normal_cand_c is None:
            normal_cand_c = nx_graph.get_vertex_candidates(stripped_gc, gs)
        if normal_cand_a is None:
            normal_cand_a = nx_graph.get_vertex_candidates(stripped_ga, gs)
        if normal_cand_b is None:
            normal_cand_b = nx_graph.get_vertex_candidates(stripped_gb, gs)
        vid_iso = {gn.node[qv]["vid"]: tv for qv, tv in zip(sorted(gn.nodes()), isomorphism)}
        v_isomorphism = set(isomorphism)
        vert_cand_c = _get_isomorphism_vertex_candidates(gc, normal_cand_c, vs, v_isomorphism, vid_iso)
        vert_cand_a = _get_isomorphism_vertex_candidates(ga, normal_cand_a, vs, v_isomorphism, vid_iso)
        vert_cand_b = _get_isomorphism_vertex_candidates(gb, normal_cand_b, vs, v_isomorphism, vid_iso)
        args_c = {"query_graph": stripped_gc, "vertex_candidates": vert_cand_c}
        args_a = {"query_graph": stripped_ga, "vertex_candidates": vert_cand_a}
        args_b = {"query_graph": stripped_gb, "vertex_candidates": vert_cand_b}
        o11, r1, c1, n, inconsistent = _frequency_signature(functools.partial(subgraph_enumeration.subsumes_nx, target_graph=gs), args_c, args_a, args_b)
        ct["o11"] += o11
        ct["r1"] += r1
        ct["c1"] += c1
        ct["n"] += n
        if inconsistent:
            ct["inconsistent"] += 1
    return ct


def subgraphs(gc, ga, gb, gn, gs, candidates=None):
    """Count subgraphs

    Arguments:
    - `go11`:
    - `gr1`:
    - `gc1`:
    - `gn`:
    - `gs`:
    - `candidates`:

    """
    ct = {x: 0 for x in ["o11", "r1", "c1", "n", "inconsistent"]}
    vid_n = set([l["vid"] for v, l in gn.nodes(data=True)])
    vs = set(gs.nodes())
    stripped_gc = strip_vid(gc)
    stripped_ga = strip_vid(ga)
    stripped_gb = strip_vid(gb)
    normal_cand_c, normal_cand_a, normal_cand_b = None, None, None
    for subgraph in subgraph_enumeration.get_subgraphs_nx(strip_vid(gn), gs, vertex_candidates=candidates):
        if normal_cand_c is None:
            normal_cand_c = nx_graph.get_vertex_candidates(stripped_gc, gs)
        if normal_cand_a is None:
            normal_cand_a = nx_graph.get_vertex_candidates(stripped_ga, gs)
        if normal_cand_b is None:
            normal_cand_b = nx_graph.get_vertex_candidates(stripped_gb, gs)
        v_subgraph = set(subgraph.nodes())
        vert_cand_c = _get_subgraph_vertex_candidates(gc, normal_cand_c, vid_n, vs, v_subgraph)
        vert_cand_a = _get_subgraph_vertex_candidates(ga, normal_cand_a, vid_n, vs, v_subgraph)
        vert_cand_b = _get_subgraph_vertex_candidates(gb, normal_cand_b, vid_n, vs, v_subgraph)
        args_c = {"query_graph": stripped_gc, "vertex_candidates": vert_cand_c}
        args_a = {"query_graph": stripped_ga, "vertex_candidates": vert_cand_a}
        args_b = {"query_graph": stripped_gb, "vertex_candidates": vert_cand_b}
        o11, r1, c1, n, inconsistent = _frequency_signature(functools.partial(subgraph_enumeration.subsumes_nx, target_graph=gs), args_c, args_a, args_b)
        ct["o11"] += o11
        ct["r1"] += r1
        ct["c1"] += c1
        ct["n"] += n
        if inconsistent:
            ct["inconsistent"] += 1
    return ct


def choke_points(gc, ga, gb, gn, gs, choke_point, candidates=None):
    """Count choke points

    Arguments:
    - `go11`:
    - `gr1`:
    - `gc1`:
    - `gn`:
    - `gs`:
    - `choke_point`:

    """
    ct = {x: 0 for x in ["o11", "r1", "c1", "n", "inconsistent"]}
    stripped_gc = strip_vid(gc)
    stripped_ga = strip_vid(ga)
    stripped_gb = strip_vid(gb)
    choke_vid = gn.node[choke_point]["vid"]
    choke_c = [v for v, l in gc.nodes(data=True) if l["vid"] == choke_vid][0]
    choke_a = [v for v, l in ga.nodes(data=True) if l["vid"] == choke_vid][0]
    choke_b = [v for v, l in gb.nodes(data=True) if l["vid"] == choke_vid][0]
    normal_cand_c, normal_cand_a, normal_cand_b = None, None, None
    for choke_point_vertex in subgraph_enumeration.get_choke_point_matches(strip_vid(gn), gs, choke_point, vertex_candidates=candidates):
        if normal_cand_c is None:
            normal_cand_c = nx_graph.get_vertex_candidates(stripped_gc, gs)
        if normal_cand_a is None:
            normal_cand_a = nx_graph.get_vertex_candidates(stripped_ga, gs)
        if normal_cand_b is None:
            normal_cand_b = nx_graph.get_vertex_candidates(stripped_gb, gs)
        args_c = {"query_graph": stripped_gc, "choke_point": choke_c, "vertex_candidates": normal_cand_c}
        args_a = {"query_graph": stripped_ga, "choke_point": choke_a, "vertex_candidates": normal_cand_a}
        args_b = {"query_graph": stripped_gb, "choke_point": choke_b, "vertex_candidates": normal_cand_b}
        o11, r1, c1, n, inconsistent = _frequency_signature(functools.partial(subgraph_enumeration.choke_point_subsumes_nx, target_graph=gs, choke_point_candidate=choke_point_vertex), args_c, args_a, args_b)
        ct["o11"] += o11
        ct["r1"] += r1
        ct["c1"] += c1
        ct["n"] += n
        if inconsistent:
            ct["inconsistent"] += 1
    return ct


def sentences(gc, ga, gb, gn, gs, candidates=None):
    """Count sentences

    Arguments:
    - `go11`:
    - `gr1`:
    - `gc1`:
    - `gn`:
    - `gs`:
    - `candidates`:

    """
    ct = {x: 0 for x in ["o11", "r1", "c1", "n", "inconsistent"]}
    subsumed_by_n = subgraph_enumeration.subsumes_nx(strip_vid(gn), gs, vertex_candidates=candidates)
    if subsumed_by_n:
        args_c = {"query_graph": strip_vid(gc)}
        args_a = {"query_graph": strip_vid(ga)}
        args_b = {"query_graph": strip_vid(gb)}
        o11, r1, c1, n, inconsistent = _frequency_signature(functools.partial(subgraph_enumeration.subsumes_nx, target_graph=gs), args_c, args_a, args_b)
        ct["o11"] += o11
        ct["r1"] += r1
        ct["c1"] += c1
        ct["n"] += n
        if inconsistent:
            ct["inconsistent"] += 1
    return ct


def run_queries(args):
    """Run all queries on a single sentence.

    Arguments:
    - `args`:

    """
    sentence, input_format, queries = args
    result = []
    if input_format == "cwb":
        gs = nx_graph.create_nx_digraph_from_cwb(sentence)
    elif input_format == "conllu":
        gs = nx_graph.create_nx_digraph_from_conllu(sentence)
    sensible = nx_graph.is_sensible_graph(gs)
    if sensible:
        for qline in queries:
            gc, ga, gb, gn, choke_point = qline
            candidates = nx_graph.get_vertex_candidates(strip_vid(gn), gs)
            # result.append(_run_query(gc, ga, gb, gn, gs, choke_point, candidates, all_counting_methods))
            result.append(_run_query_all(gc, ga, gb, gn, gs, choke_point, candidates))
    return result, sensible


def run_queries_db(args):
    """Run a query on a single sentence from the database.

    Arguments:
    - `args`:

    """
    gc, ga, gb, gn, choke_point, sentence = args
    gs = json_graph.node_link_graph(json.loads(sentence))
    candidates = nx_graph.get_vertex_candidates(strip_vid(gn), gs)
    # return _run_query(gc, ga, gb, gn, gs, choke_point, candidates)
    return _run_query_all(gc, ga, gb, gn, gs, choke_point, candidates)


def _run_query(gc, ga, gb, gn, gs, choke_point, candidates):
    """Run a single query on a single graph"""
    result = {}
    # choke_points (contingency table)
    choke_point_ct = {}
    if choke_point is not None:
        choke_point_ct = choke_points(gc, ga, gb, gn, gs, choke_point, candidates)
        result["focus_points"] = choke_point_ct
    # isomorphisms
    iso_ct = isomorphisms(gc, ga, gb, gn, gs, candidates)
    result["embeddings"] = iso_ct
    # subgraphs (contingency table)
    sub_ct = subgraphs(gc, ga, gb, gn, gs, candidates)
    result["subgraphs"] = sub_ct
    # sentences (contingency table)
    sent_ct = sentences(gc, ga, gb, gn, gs, candidates)
    result["sentences"] = sent_ct
    return result


def _run_query_all(gc, ga, gb, gn, gs, choke_point, candidates):
    """Run a single query on a single graph"""
    return all_counting_methods(gc, ga, gb, gn, gs, choke_point, candidates)


def merge_result(result, results):
    """Merge result with results

    Arguments:
    - `result`:
    - `results`:
    """
    for i, query in enumerate(result):
        for method in query:
            if method not in results[i]:
                results[i][method] = {}
            for frequency in query[method]:
                results[i][method][frequency] = results[i][method].get(frequency, 0) + query[method][frequency]


def merge_result_db(result, query_number, results):
    """Merge result with results

    Arguments:
    - `query`:
    - `query_number`:
    - `results`:
    """
    for method in result:
        if method not in results[query_number]:
            results[query_number][method] = {}
        for frequency in result[method]:
            results[query_number][method][frequency] = results[query_number][method].get(frequency, 0) + result[method][frequency]



# 1. Upload to PyPI:
# python3 setup.py sdist
# python3 setup.py sdist upload
#
# 2. Check if everything looks all right: https://pypi.python.org/pypi/Pareidoscope
#
# 3. Create tag and push to Github:
# git tag -a v<version> -m "annotation for this release"
# git push origin --tags

from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as fh:
    long_description = fh.read()

version = "0.9.4"

setup(
    name='Pareidoscope',
    version=version,
    author='Thomas Proisl',
    author_email='thomas.proisl@fau.de',
    packages=[
        'pareidoscope',
        'pareidoscope.utils'
    ],
    scripts=[
        'bin/pareidoscope_associated_structures',
        'bin/pareidoscope_association_strength',
        'bin/pareidoscope_collexeme_analysis',
        'bin/pareidoscope_corpus_to_sqlite',
        'bin/pareidoscope_covarying_collexemes',
        'bin/pareidoscope_draw_graphs',
    ],
    url='https://github.com/tsproisl/Pareidoscope',
    download_url='https://github.com/tsproisl/Pareidoscope/archive/v%s.tar.gz' % version,
    # include_package_data=True,
    # url='http://pypi.python.org/pypi/Pareidoscope/',
    license='GNU General Public License v3 or later (GPLv3+)',
    description='A collection of tools for determining the association between arbitrary linguistic structures.',
    long_description=long_description,
    install_requires=[
        'networkx >= 1.6',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Text Processing :: Linguistic',
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re


def sentences_iter(corpus, return_id=False):
    """Iterate over the sentences in a corpus.

    Arguments:
    - `corpus`:
    """
    pattern = re.compile(r"\bid=(['\"])([^'\"]+)\1")
    origid = ""
    sentence = []
    for line in corpus:
        line = line.rstrip("\n")
        if line == "</s>":
            if return_id:
                yield sentence, origid
            else:
                yield sentence
        elif line.startswith("<s "):
            m = re.search(pattern, line)
            if m:
                origid = m.group(2)
            else:
                raise Exception("Line does not match: %s" % line)
            sentence = []
        elif line.startswith("<"):
            pass
        else:
            sentence.append(line.split("\t"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import re
import sqlite3


def create_db(filename):
    """Connect to database, create tables and indexes and return
    connection and cursor."""
    conn = sqlite3.connect(filename)
    c = conn.cursor()
    c.execute("PRAGMA page_size=4096")
    c.execute("PRAGMA cache_size=100000")
    c.execute("CREATE TABLE sentences (sentence_id INTEGER PRIMARY KEY AUTOINCREMENT, original_id TEXT, graph TEXT, UNIQUE (original_id))")
    c.execute("CREATE TABLE tokens (token_id INTEGER PRIMARY KEY AUTOINCREMENT, sentence_id INTEGER, position INTEGER, word TEXT, pos TEXT, lemma TEXT, wc TEXT, indegree INTEGER, outdegree INTEGER, root BOOLEAN, FOREIGN KEY (sentence_id) REFERENCES sentences (sentence_id), UNIQUE (sentence_id, position))")
    c.execute("CREATE TABLE dependencies (governor_id INTEGER, dependent_id INTEGER, relation TEXT, FOREIGN KEY (governor_id) REFERENCES tokens (token_id), FOREIGN KEY (dependent_id) REFERENCES tokens (token_id), UNIQUE (governor_id, dependent_id))")
    c.execute("CREATE INDEX word_idx ON tokens (word)")
    c.execute("CREATE INDEX pos_idx ON tokens (pos)")
    c.execute("CREATE INDEX lemma_idx ON tokens (lemma)")
    c.execute("CREATE INDEX wc_idx ON tokens (wc)")
    c.execute("CREATE INDEX indegree_idx ON tokens (indegree)")
    c.execute("CREATE INDEX outdegree_idx ON tokens (outdegree)")
    c.execute("CREATE INDEX root_idx ON tokens (root)")
    c.execute("CREATE INDEX governor_id_relation_idx ON dependencies (governor_id, relation)")
    c.execute("CREATE INDEX dependent_id_relation_idx ON dependencies (dependent_id, relation)")
    c.execute("CREATE INDEX relation_idx ON dependencies (relation)")
    return conn, c


def insert_sentence(c, origid, gs, graph):
    """Insert sentence into database"""
    c.execute("INSERT INTO sentences (original_id, graph) VALUES (?, ?)", (origid, graph))
    sid = c.execute("SELECT sentence_id FROM sentences WHERE original_id = ?", (origid, )).fetchall()[0][0]
    token_id = {}
    for vertice in sorted(gs.nodes()):
        word = gs.node[vertice]["word"]
        pos = gs.node[vertice]["pos"]
        lemma = gs.node[vertice]["lemma"]
        wc = gs.node[vertice]["wc"]
        root = "root" in gs.node[vertice] and gs.node[vertice]["root"] == "root"
        c.execute("INSERT INTO tokens (sentence_id, position, word, pos, lemma, wc, indegree, outdegree, root) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)", (sid, vertice, word, pos, lemma, wc, gs.in_degree(vertice), gs.out_degree(vertice), root))
        token_id[vertice] = c.execute("SELECT token_id FROM tokens WHERE sentence_id = ? AND position = ?", (sid, vertice)).fetchall()[0][0]
    for s, t, l in gs.edges(data=True):
        c.execute("INSERT INTO dependencies (governor_id, dependent_id, relation) VALUES (?, ?, ?)", (token_id[s], token_id[t], l["relation"]))


def connect_to_database(filename, re=False):
    """Connect to database and return connection and cursor."""
    conn = sqlite3.connect(filename)
    # For the PCRE extension:
    # conn.enable_load_extension(True)
    # conn.load_extension("/usr/lib/sqlite3/pcre.so")
    c = conn.cursor()
    c.execute("PRAGMA page_size=4096")
    c.execute("PRAGMA cache_size=100000")
    if re:
        conn.create_function("REGEXP", 2, regexp)
    return conn, c


def sentence_candidates(c, g):
    """Get candidate sentences (no token candidates!) for the query
    graph.

    """
    sentence_ids = []
    for vertex in g.nodes():
        sql_query = "SELECT DISTINCT sentence_id FROM tokens WHERE "
        where = []
        args = []
        pos_lexical = set(["word", "pos", "lemma", "wc", "root"])
        neg_lexical = set(["not_%s" % pl for pl in pos_lexical])
        indegree = g.in_degree(vertex)
        outdegree = g.out_degree(vertex)
        if indegree > 0:
            where.append("indegree >= ?")
            args.append(indegree)
        if outdegree > 0:
            where.append("outdegree >= ?")
            args.append(outdegree)
        for k, v in g.node[vertex].items():
            if k in pos_lexical:
                where.append("%s = ?" % k)
                if k == "root":
                    args.append(v == "root")
                else:
                    args.append(v)
            elif k in neg_lexical:
                k = k[4:]
                where.append("%s != ?" % k)
                if k == "root":
                    args.append(v == "root")
                else:
                    args.append(v)
            elif k == "not_indep":
                relations = []
                for rel in v:
                    relations.append("relation = ?")
                    args.append(rel)
                where.append("NOT EXISTS (SELECT 1 FROM dependencies WHERE dependent_id = token_id AND (%s))" % " OR ".join(relations))
            elif k == "not_outdep":
                relations = []
                for rel in v:
                    relations.append("relation = ?")
                    args.append(rel)
                where.append("NOT EXISTS (SELECT 1 FROM dependencies WHERE governor_id = token_id AND (%s))" % " OR ".join(relations))
            else:
                raise Exception("Unsupported key: %s" % k)
        sql_query += " AND ".join(where)
        sentence_ids.append(set([r[0] for r in c.execute(sql_query, args).fetchall()]))
    candidate_ids = functools.reduce(lambda x, y: x.intersection(y), sentence_ids)
    candidate_sentences = (r[0] for r in c.execute("SELECT graph FROM sentences WHERE sentence_id IN (%s)" % ", ".join([str(_) for _ in candidate_ids])))
    return candidate_sentences


def regexp(expression, item):
    """Does expression match item?

    Arguments:
    - `expression`:
    - `item`:

    """
    reg = re.compile(expression)
    return reg.search(item) is not None


# Only used in pareidoscope_collexeme_analysis_db
def create_sql_query(query_graph):
    """Create an SQL query for the given query_graph."""
    sql_query = "SELECT s.sentence_id, s.graph, "
    where = []
    arguments = []
    sql_query += ", ".join(["tok_%s.position" % v for v in query_graph.nodes()])
    sql_query += " FROM sentences AS s"
    for vertex in query_graph.nodes():
        sql_query += " INNER JOIN tokens AS tok_%s ON s.sentence_id = tok_%s.sentence_id" % (vertex, vertex)
    for s, t, l in query_graph.edges(data=True):
        sql_query += " INNER JOIN dependencies AS dep_%s_%s ON (dep_%s_%s.governor_id = tok_%s.token_id) AND (dep_%s_%s.dependent_id = tok_%s.token_id)" % (s, t, s, t, s, s, t, t)
    sql_query += " WHERE "
    pos_lexical = set(["word", "pos", "lemma", "wc", "root"])
    neg_lexical = set(["not_%s" % pl for pl in pos_lexical])
    for vertex in query_graph.nodes():
        indegree = query_graph.in_degree(vertex)
        outdegree = query_graph.out_degree(vertex)
        if indegree > 0:
            where.append("tok_%s.indegree >= %d" % (vertex, indegree))
        if outdegree > 0:
            where.append("tok_%s.outdegree >= %d" % (vertex, outdegree))
        for k, v in query_graph.node[vertex].items():
            if k in pos_lexical:
                where.append("tok_%s.%s = ?" % (vertex, k))
                if k == "root":
                    arguments.append(v == "root")
                else:
                    arguments.append(v)
            elif k in neg_lexical:
                where.append("tok_%s.%s != ?" % (vertex, k))
                if k == "root":
                    arguments.append(v == "root")
                else:
                    arguments.append(v)
            elif k == "not_indep":
                for rel_item in v:
                    for rel in rel_item.split("|"):
                        where.append("? NOT IN (SELECT relation FROM dependencies WHERE dependent_id = tok_%s.token_id)" % vertex)
                        arguments.append(rel)
            elif k == "not_outdep":
                for rel_item in v:
                    for rel in rel_item.split("|"):
                        where.append("? NOT IN (SELECT relation FROM dependencies WHERE governor_id = tok_%s.token_id)" % vertex)
                        arguments.append(rel)
            else:
                raise Exception("Unsupported key: %s" % k)
    for s, t, l in query_graph.edges(data=True):
        if "relation" in l:
            where.append("dep_%s_%s.relation = ?" % (s, t))
            arguments.append(l["relation"])
    sql_query += " AND ".join(where)
    return sql_query, tuple(arguments)


# Only used in pareidoscope_collexeme_analysis_db
def aggregate_sentences(c, query, query_args):
    """Aggregate sentences, returning sentence id, sentence graph, and
    sets of candidates."""
    old_sentence_id = None
    candidates = None
    old_graph = None
    for row in c.execute(query, query_args):
        sentence_id = row[0]
        graph = row[1]
        token_positions = row[2:]
        if sentence_id != old_sentence_id and old_sentence_id is not None:
            yield old_sentence_id, old_graph, candidates
            candidates = None
        if candidates is None:
            candidates = [set([t]) for t in token_positions]
        else:
            for i, t in enumerate(token_positions):
                candidates[i].add(t)
        old_sentence_id = sentence_id
        old_graph = graph



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sqlite3

def _connect_to_db(filename):
    """"""
    dirname = os.path.dirname(filename)
    conn = sqlite3.connect(filename)
    c = conn.cursor()
    c.execute("PRAGMA page_size=4096")
    c.execute("PRAGMA cache_size=100000")
    c.execute("PRAGMA temp_store=1")
    c.execute("PRAGMA temp_store_directory='%s'" % dirname)
    c.execute("CREATE TABLE IF NOT EXISTS queries (queryid INTEGER PRIMARY KEY AUTOINCREMENT, query TEXT, islemma BOOLEAN, collexeme_index INTEGER, wordclass TEXT, UNIQUE (query, islemma, collexeme_index, wordclass))")
    c.execute("CREATE TABLE IF NOT EXISTS associations (queryid INTEGER, collexeme TEXT, counting_method TEXT, dice REAL, gmean REAL, t_score REAL, log_likelihood REAL, o11 INTEGER, r1 INTEGER, c1 INTEGER, n INTEGER, UNIQUE(queryid, collexeme, counting_method), FOREIGN KEY (queryid) REFERENCES queries)")
    c.execute("CREATE TABLE IF NOT EXISTS correlations (queryid INTEGER, counting_method_1 TEXT, counting_method_2 TEXT, rho_dice REAL, rho_gmean REAL, rho_t_score REAL, rho_log_likelihood REAL, UNIQUE (queryid, counting_method_1, counting_method_2), FOREIGN KEY (queryid) REFERENCES queries)")
    return conn, c


def insert_results(db_filename, query, islemma, collexeme_index, wordclass, o11, r1, c1, n, dice_associations, gmean_associations, tscore_associations, ll_associations, dice_correlations, gmean_correlations, tscore_correlations, ll_correlations):
    """"""
    conn, c = _connect_to_db(db_filename)
    c.execute("INSERT OR IGNORE INTO queries (query, islemma, collexeme_index, wordclass) VALUES (?, ?, ?, ?)", (query, islemma, collexeme_index, wordclass))
    queryid = c.execute("SELECT queryid FROM queries WHERE query=? AND islemma=? AND collexeme_index=? AND wordclass=?", (query, islemma, collexeme_index, wordclass)).fetchall()[0][0]
    counting_methods = ["collostructional", "choke_points", "subgraphs", "isomorphisms"]
    get_freq = lambda freq, cm, starid: freq[cm] if cm == "collostructional" else freq[cm][starid]
    association_tuples = ((queryid, collexeme, counting_method, dice_associations[collexeme][counting_method], gmean_associations[collexeme][counting_method], tscore_associations[collexeme][counting_method], ll_associations[collexeme][counting_method], o11[counting_method][collexeme], r1[counting_method][collexeme], c1[counting_method], n[counting_method]) for collexeme in ll_associations.keys() for counting_method in counting_methods)
    c.executemany("INSERT OR IGNORE INTO associations (queryid, collexeme, counting_method, dice, gmean, t_score, log_likelihood, o11, r1, c1, n) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", association_tuples)
    correlation_tuples = ((queryid, cm1, cm2, dice_correlations[i][j], gmean_correlations[i][j], tscore_correlations[i][j], ll_correlations[i][j]) for i, cm1 in enumerate(counting_methods) for j, cm2 in enumerate(counting_methods))
    c.executemany("INSERT OR IGNORE INTO correlations (queryid, counting_method_1, counting_method_2, rho_dice, rho_gmean, rho_t_score, rho_log_likelihood) VALUES (?, ?, ?, ?, ?, ?, ?)", correlation_tuples)
    conn.commit()
    conn.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re


def sentences_iter(corpus, return_id=False):
    """Iterate over the sentences in a corpus in CoNLL-U format
    (http://universaldependencies.org/format.html).

    Arguments:
    - `corpus`:

    """
    pattern = re.compile(r"^#\s*sent_id\s*=\s*(\S.*)$")
    origid = ""
    sentence = []
    for line in corpus:
        line = line.rstrip()
        if line == "":
            if return_id:
                yield sentence, origid
            else:
                yield sentence
            sentence = []
        elif line.startswith("#") and len(sentence) == 0:
            m = re.search(pattern, line)
            if m:
                origid = m.group(1)
        else:
            sentence.append(line.split("\t"))
